# -*- cperl -*-

BEGIN
{
	import PML_T_Edit;
}

=item detect()

Returns true if the current file is an English PDT t-file

=cut

sub detect {
  return (PML::SchemaDescription() =~ m{^English PDT} and PML::SchemaName() eq 'tdata') ? 1 : 0;
}

sub allow_switch_context_hook {
  return 'stop' unless detect();
}


##### Layer switching ----------------------------

sub AnalyticalTree {
  return unless SchemaName() eq 'tdata';
  return unless SwitchToAFile();
  my $t_root = $root;
  my $t_node = $this;
  $PML_En_T::last_context = CurrentContext();
  SwitchContext('PML_En_A');
  my $fsfile = $grp->{FSFile};
  #find current tree and new $this
  my $trees = $fsfile->treeList;
  my $treeNo = $fsfile->currentTreeNo+0;
 TREE: for (my $i=0;$i<=$#$trees;$i++) {
    foreach my $a_rf (GetANodeIDs($t_root)) {
      $a_rf =~ s/^.*?#//;
      if ($trees->[$i]->{id} eq $a_rf) {
	$treeNo = $i;
	last TREE;
      }
    }
  }
  TredMacro::GotoTree($treeNo+1); # changes $root and $this
  my $a_ids = GetNodeHash($fsfile);
  my $first =
    first {
      ref($a_ids->{$_}) and ($a_ids->{$_}->root == $root) ? 1 : 0
    } GetANodeIDs($t_node);
  $this = $a_ids->{$first};
  ChangingFile(0);
  return 1;
}

sub PhrasalTree {
  AnalyticalTree();
  PML_En_A::PhrasalTree();
}

sub MarkForARf {
  $PML::arf=$this;
  ChangingFile(0);
  PML_En_T::AnalyticalTree();
}

##### EngValLex ----------------------------------

my $engvallex;
my $prefix;

sub init_engvallex
{
	return if $engvallex;
	my $file;
	($prefix,$file) = FindVallex();
	$engvallex = TrEd::EngValLex::NonGUI->new($file);
}

sub _get_frame {
  my$rf=shift;
  init_engvallex();
  $rf=~s/^\Q$prefix\E\#//;
  my $frame = $engvallex->by_id($rf);
  return $engvallex->serialize_frame($frame) if $frame;
  'FRAME NOT FOUND!';
}

sub OpenValFrameList {
  shift unless @_ and ref($_[0]);
  my $node = shift || $this;
  my %opts = @_;

  local $EngValLex::GUI::frameid_attr="val_frame.rf";
  local $EngValLex::GUI::lemma_attr="t_lemma";
  local $EngValLex::GUI::framere_attr=undef;
  local $EngValLex::GUI::sempos_attr="gram/sempos";
  $TrEd::EngValLex::Editor::reviewer_can_modify = 1;
  $TrEd::EngValLex::Editor::reviewer_can_delete = 1;

  my $a_node = GetANodeByID($node->attr('a/lex.rf'));

  my ($refid, $vallex_file) = FindVallex();
  my $rf = $node ? join('|',map { my $x=$_;$x=~s/^\Q$refid\E#//; $x } AltV($node->{'val_frame.rf'})) : undef;
  TrEd::EngValLex::GUI::ChooseFrame({
	-vallex_file => $vallex_file,
    -lemma => $node ? $node->{t_lemma} : undef,
	-sempos => $node ? $node->attr('gram/sempos') : undef,
	-lemma_attr => 't_lemma',
	-sempos_attr => 'gram/sempos',
    -frameid => $rf,
    -assignfunc => sub{},
	-pos => $a_node->attr('m/tag'),
    %opts
  });
  ChangingFile(0);
}

sub ChooseValFrame {
  shift unless @_ and ref($_[0]);
  my $node = shift || $this;
  my %opts = @_;
  my $refid = FileMetaData('refnames')->{vallex};

  my $fsfile = $grp->{FSFile};
  OpenValFrameList(
    $node,
	-withdraw => 1,
    -assign_func => sub {
      my ($n, $ids, undef, $current_fsfile, $win, $chooser)=@_;
	  $grp = $win->{focusedWindow};
	  if ($current_fsfile == $fsfile)
	  {
		  AssignValencyFrames($n, split /\|/, $ids);	#included from PML_T_Edit
	  }
	  else
	  {
		  my $top = $chooser ? $chooser->widget->toplevel : TopLevelFrame();
		  $top->messageBox(-icon => 'error', -message => 'Use Ctrl + Return to assign frame!', -title => 'Error', -type => 'ok');
	  }
    },
    %opts
   );
  ChangingFile(0);
}

sub OpenValLexicon {
  shift unless @_ and ref($_[0]);
  my $node = shift || $this;
  my %opts = @_;

  local $TrEd::EngValLex::GUI::frameid_attr="val_frame.rf";
  local $TrEd::EngValLex::GUI::lemma_attr="t_lemma";
  local $TrEd::EngValLex::GUI::framere_attr=undef;
  local $TrEd::EngValLex::GUI::sempos_attr="gram/sempos";
  $TrEd::EngValLex::Editor::reviewer_can_modify = 1;
  $TrEd::EngValLex::Editor::reviewer_can_delete = 1;

  my $a_node = GetANodeByID($node->attr('a/lex.rf'));

  my ($refid, $vallex_file) = FindVallex();
  my $rf = $node ? join('|',map { my $x=$_;$x=~s/^\Q$refid\E#//; $x } AltV($node->{'val_frame.rf'})) : undef;
  TrEd::EngValLex::GUI::OpenEditor(
  {
	-vallex_file => $vallex_file,
    -lemma => $node ? $node->{t_lemma} : undef,
    -sempos => $node ? $node->attr('gram/sempos') : undef,
    -frameid => $rf,
	-pos => $a_node->attr('m/tag'),
    %opts
  });
  ChangingFile(0);
  return;
}

##### Hooks --------------------------------------

sub after_edit_attr_hook {
  my($node,$attr,$result)=(shift,shift,shift);
  return unless $result;
  if($attr eq 't_lemma'){
    if($node->{t_lemma}=~m/^#(.*)/
       and not first{$1 eq $_}qw[Amp Ast AsMuch Benef Bracket Colon
                                 Comma Cor Dash EmpNoun EmpVerb Equal
                                 Forn Gen Idph Neg Oblfm Percnt
                                 PersPron Period Period3 QCor Rcp
                                 Slash Separ Some Total Unsp VerbPron]){
      questionQuery('Invalid entity',
                    'Cannot change t-lemma to undefined #-entity. T-lemma not changed.',
                    'OK');
      Undo();
    }else{
      VallexWarning($node);
    }
  }
  if ($attr =~ m/^(?:t_lemma|functor)$/)
  {
  	assign_correct_nodetype($node);
  }
  if (($attr eq 'functor')&&($result eq 'PAR')){
  $node->{'is_parenthesis'} = 1;
  }
}

sub allow_switch_context_hook {
  return 'stop' if SchemaName() ne 'tdata';
}

sub enable_attr_hook
{
  if ($_[0] =~ m#^(id|a|a/.*|compl\.rf.*|coref_gram\.rf.*|coref_text\.rf.*|deepord|quot|val_frame\.rf|bbn_tag|nombank_data|highlite|nodetype|functor_change)$#)
    {'stop'} else {1}
}

sub enable_edit_node_hook { 1 }

sub get_status_line_hook {
  return unless $this;
  my$statusline= [
	  # status line field definitions ( field-text => [ field-styles ] )
	  [
#	   "     id: " => [qw(label)],
#	   $this->{id} => [qw({id} value)],
	   "     a:" => [qw(label)]
          ],
	  # field styles
	  [
           "ref" => [-underline => 1, -foreground => 'blue'],
	   "label" => [-foreground => 'black' ],
	   "value" => [-underline => 1 ],
	   "bg_white" => [ -background => 'white' ],
           "status" => [ -foreground => CustomColor('status')]
	  ]
	 ];
  my$sep=" ";
  foreach my $ref (
                   $this->{nodetype}eq'root'
                   ?
                   $this->{'atree.rf'}
                   :
                   ($this->attr('a/lex.rf'),ListV($this->attr('a/aux.rf')))){
    if($ref){
      push @{$statusline->[0]},
        ($sep => [qw(label)],GetANodeByID($ref)->attr('m/form') => [ '{REF}','ref',$ref ]);
    $sep=", ";
    }
  }
  unshift @{$statusline->[0]},
    ($this->{'val_frame.rf'} ?
     ("     frame: " => [qw(label)],
      join(",",map{_get_frame($_)}AltV($this->{'val_frame.rf'})) => [qw({FRAME} value)]
     ) : ());
  push @{$statusline->[0]},
    ($PML_T_Edit::remember ?
     ('   Remembered: ' => [qw(label)],
      $PML_T_Edit::remember->{t_lemma} || $PML_T_Edit::remember->{id}=> [qw(status)]
     ):'');
  return $statusline;
}#get_status_line_hook

sub get_value_line_hook {
  my ($fsfile,$treeNo) = @_;
  return unless $fsfile;
  my $tree = $fsfile->tree($treeNo);
  return unless $tree;
  my @out = @{&PML_T::get_value_line_hook};
  if($tree->{cze_sentence}){
    push @out,(["\n".$tree->{cze_sentence},'-foreground=>blue']);
  }
  return \@out;
}

sub node_release_hook{
  &PML_T_Edit::node_release_hook;
  my ($node,$target,$mod)=@_;
  return if $mod || ! $target;
  if ($target->{t_lemma} eq "#Forn") {
    $node->{functor}='FPHR';
    $node->{nodetype}='fphr';
    TredMacro::Redraw_FSFile_Tree();
    $FileChanged=1;
  }
}

sub status_line_doubleclick_hook {
  # status-line field double clicked

  # @_ contains a list of style names associated with the clicked
  # field. Style names may obey arbitrary user-defined convention.

  foreach (@_) {
    if (/^\{(.*)}$/) {
      if ($1 eq 'FRAME') {
	    PML_En_T_Edit::ChooseValFrame();
        last;
      } elsif($1 eq 'REF'){
        my $aref=$_[-1];
        $aref=~s/.*?#//;
        AnalyticalTree();
        $this=$root;
        my($node,$tree)=SearchForNodeById($aref);
        TredMacro::GotoTree($tree);
        $this=$node;
        Redraw_FSFile();
      } else {
        if (EditAttribute($grp,$this,$1)) {
          ChangingFile(1);
          Redraw_FSFile();
        }
        last;
      }
    }
  }
}

sub switch_context_hook {
  my $cur_stylesheet = TredMacro::GetCurrentStylesheet();
  if ($cur_stylesheet ne 'PEDT_Anot'){
    SetCurrentStylesheet('PEDT_Anot');
    Redraw() if GUI();
  }
  undef $PML::arf;
  return;
}

sub value_line_click_hook {
  PML_T_Edit::MouseEditLinks(0,@_);
}

##### Edit 1/0 attributes ------------------------

##vynechana makra
sub RotateIsName {
  $this->{is_name}=!$this->{is_name};
}

sub RotateIsNameSubtree{
  my $name = ! $this->{is_name};
  foreach my $node ($this,$this->descendants) {
    $node->{is_name} = $name;
  }
}

sub RotateIsNameOfPerson {
  $this->{is_name_of_person}=!$this->{is_name_of_person};
  $this->{is_name} = $this->{is_name_of_person};
}

sub RotateIsNameOfPersonSubtree{
  my $name = ! $this->{is_name_of_person};
  foreach my $node ($this,$this->descendants) {
    $node->{is_name_of_person} = $node->{is_name} = $name;
  }
}
## konec vynechanych maker

sub RotateState{
    $this->{is_state}=!$this->{is_state};
}

sub RotateDsp{
    $this->{is_dsp_root}=!$this->{is_dsp_root};
}

sub RotateParenthesisSubtree{
  my $par = ! $this->{is_parenthesis};
  foreach my $node ($this,$this->descendants) {
    $node->{is_parenthesis} = $par;
  }
}

##### Edit other attributes ----------------------

sub EditComment{
  ChangingFile(EditAttribute($this,'annot_comment'));
}

sub EditSubfunctor{
  ChangingFile(EditAttribute($this,'subfunctor'));
}

sub EditFunctor{
  ChangingFile(EditAttribute($this,'functor'));
}

sub EditSemPOS{
  ChangingFile(EditAttribute($this,'gram/sempos'));
}

sub EditTLemma{
  ChangingFile(EditAttribute($this,'t_lemma'));
  # Because of Undo in hook:
  $this=undef;
}

sub EditVerbPatt{
  my $was_defined = defined($this->{pcedt});
  $this->{pcedt} = Treex::PML::Factory->createStructure unless $was_defined;

  if (EditAttribute($this, 'pcedt')){
	ChangingFile(1);
  }
  else{
	ChangingFile(0);
	delete($this->{pcedt}) unless $was_defined;
  }
}

##### Create new nodes ---------------------------

# Overload AddENode and AddEntityNode because of VerbPron
sub AddENode {
  ChangingFile(0);
  AddEntityNode(1);
}

sub AddEntityNode {
  my $autotype=shift;
  my$dialog=[];
  my%lemmas=qw/AsMuch qcomplex
               Benef qcomplex
               Colon coap
               Comma coap
               Cor qcomplex
               Dash coap
               EmpNoun complex
               EmpVerb qcomplex
               Equal qcomplex
               Forn list
               Gen qcomplex
               Idph list
               Neg atom
               Oblfm qcomplex
               Percnt qcomplex
               PersPron complex
               QCor qcomplex
               Rcp qcomplex
               Separ coap
               Some qcomplex
               Total qcomplex
               Unsp qcomplex
               VerbPron complex
               /;
  $lemmas{NewNode}='qcomplex'if$autotype;
  ListQuery('T-lemma',
            'browse',
            [sort keys%lemmas],
            $dialog) or return;
  $this=NewNode($this);
  $this->{functor}='PAR';
  $this->{t_lemma}='#'.$dialog->[0];
  $this->{nodetype}=$lemmas{$dialog->[0]};
  $this->{is_generated}=1;
  #EditFunctor()|| (DeleteNode() && return );
  unless ($autotype){
    EditNodetype()|| (DeleteNode() && return );
  }
}#AddEntityNode

sub AddNeg {
  $this=NewNode($this);
  $this->{functor}='RHEM';
  $this->{t_lemma}='#Neg';
  $this->{nodetype}='atom';
  $this->{is_generated}=1;
}

sub AddEmpVerb {
  $this=NewNode($this);
  $this->{functor}='PRED';
  $this->{t_lemma}='#EmpVerb';
  $this->{nodetype}='qcomplex';
  $this->{is_generated}=1;
}

sub AddEmpNoun {
  $this=NewNode($this);
  $this->{functor}='VOCAT';
  $this->{t_lemma}='#EmpNoun';
  $this->{nodetype}='qcomplex';
  $this->{is_generated}=1;
}

sub AddForn {
  $this=NewNode($this);
  $this->{functor}='ID';
  $this->{t_lemma}='#Forn';
  $this->{nodetype}='list';
  $this->{is_generated}=1;
}

sub AddNewNode {
  $this=NewNode($this);
  $this->{functor}='VOCAT';
  $this->{t_lemma}='#NewNode';
  $this->{nodetype}='complex';
  $this->{is_generated}=1;
}

sub AddAnalyticNode {
  my $autotype=shift;
  $this=NewNode($this);
  $this->{t_lemma}='#NewNode';
  $this->{functor}='???';
  $this->{nodetype}='complex';
  $this->{is_generated}=0;
  ChangingFile(1);
  MarkForARf();
}

sub AddANode {
  ChangingFile(0);
  AddAnalyticNode(1);
}

##### Creating t-lemma ---------------------------

sub _Perm{
  my$pos=shift;
  my$perm=shift;
  if($pos<@_){
    for(my$i=$pos;$i<@_;$i++){
      if($i!=$pos){
        my$j=$_[$i];
        $_[$i]=$_[$pos];
        $_[$pos]=$j;
      }
      _Perm($pos+1,$perm,@_);
      if($i!=$pos){
        my$j=$_[$i];
        $_[$i]=$_[$pos];
        $_[$pos]=$j;
      }
    }
  }else{
    push @{$perm},join"_",@_;
  };
}

sub _GenerateLemmaList{
  my$perm=[];
  _Perm(0,$perm,@_);
  return$perm;
}

sub RegenerateTLemma{
  ChangingFile(0);
  my @anodes=PML_T::GetANodes($this);
  if(@anodes>1){
    my@words=map{
      my $l = $_->attr('m/lemma');
      $l =~ s/(.+?)[-_`].*$/$1/;
      $l;
    }@anodes;
    my$d=[$words[0]];
    ListQuery('Participating lemmas',
              'multiple',
              \@words,
              $d);
    my$possible=_GenerateLemmaList(@{$d});
    $d=[$possible->[0]];
    ListQuery('New t-lemma','browse',
              $possible,
              $d) or return;
    $this->{t_lemma}=$d->[0];
    ChangingFile(1);
    VallexWarning($this);
  }elsif(@anodes==1){
    my $l = $anodes[0]->attr('m/lemma');
    $l =~ s/(.+?)[-_`].*$/$1/;
    $l =~ s/^%$/#Percnt/;
    $this->{t_lemma} = $l;
    ChangingFile(1);
    VallexWarning($this);
  }
}#RegenerateTLemma

sub TLemmaFromForms {
  ChangingFile(0);
  my @anodes=PML_T::GetANodes($this);
  SortByOrd(\@anodes);
  my $newlemma = join '_', map { $_->attr('m/form') } @anodes;
  if ($newlemma ne $this->{t_lemma}){
    $this->{t_lemma}=$newlemma;
    ChangingFile(1);
    VallexWarning($this);
  }
}

sub AssignTLemma {
  $this->{t_lemma}=$_[0];
  after_edit_attr_hook($this,'t_lemma',$_[0]);
}

sub TLemmaCor {
  AssignTLemma('#Cor');
}

sub TLemmaPersPron {
  AssignTLemma('#PersPron');
}

sub TLemmaGen {
  AssignTLemma('#Gen');
}

sub TLemmaUnsp {
  AssignTLemma('#Unsp');
}

sub TLemmaVerbPron {
  AssignTLemma('#VerbPron');
}

sub TLemmaOblfm {
	AssignTLemma('#Oblfm');
}

##### Misc. --------------------------------------

sub AddComment {
  my @list=$this->type('annot_comment/type/')->get_values();
  my$dialog=[$list[0]];
  ListQuery('Comment type',
            'browse',
            \@list,
            $dialog) or return;
  my$text=QueryString('Comment text','Text:');
  return unless defined $text;
  my%comment=(type=>$dialog->[0],
              text=>$text);
  AddToList($this,'annot_comment',\%comment);
}

sub VallexWarning {
  if($_[0]->{'val_frame.rf'}){
    questionQuery('T-lemma changed',
                  'T-lemma has changed. Verify that the vallex reference is correct.',
                  'OK');
  }
}

sub assign_correct_nodetype
{
	my $node = shift;

	if ($node->{functor} =~ m/^(?:ATT|CM|INTF|MOD|PARTL|PREC|RHEM)$/)
	{
		$node->{nodetype} = 'atom';
		return;
	}
	if ($node->{functor} =~ m/^(?:CONJ|ADVS|CSQ|DISJ|GRAD|REAS|CONFR|CONTRA|OPER|APPS)$/)
	{
		$node->{nodetype} = 'coap';
		return;
	}
	if ($node->{t_lemma} =~ m/^#(?:Idph|Forn)$/)
	{
		$node->{nodetype} = 'list';
		return;
	}
	if ($node->{functor} =~ m/^[FD]PHR$/)
	{
		$node->{nodetype} = lc $node->{functor};
		return;
	}
	if ($node->{t_lemma} =~ m/^#(?:AsMuch|Cor|EmpVerb|Equal|Gen|Oblfm|QCor|Rcp|Some|Total|Unsp|Amp|Ast|Percnt|Bracket|Comma|Colon|Dash|Period|Period3|Slash)$/)
	{
		$node->{nodetype} = 'qcomplex';
		return;
	}
	
	$node->{nodetype} = 'complex';
}

sub DeleteNodeToAux{
  shift unless ref $_[0];
  my$node=$_[0]||$this;
  ChangingFile(0),return unless $node->parent;
  my$parent=$node->parent;
  foreach my$child($node->children){
    CutPaste($child,$parent);
  }
  AddToList($parent,
            'a/aux.rf',
            grep { defined $_ }
            $node->attr('a/lex.rf'),ListV($node->attr('a/aux.rf'))
           ) if $node->attr('a/lex.rf') or $node->attr('a/aux.rf');
  DeleteLeafNode($node);
  $this=$parent unless@_;
  ChangingFile(1);
}

sub remove_all_from_aux
{
	$this->{a}->{'aux.rf'} = ();
}

sub Toggle3rdLine {
  $PML_En_T::thirdLine = ($PML_En_T::thirdLine + 1) % 3;
  ChangingFile(0);
}

sub ClearValframe {
  undef $this->{'val_frame.rf'};
}

##### Functors -----------------------------------

sub AssignFunctor {
  PML_En_T_SE::add_functor_comment($this, $_[0]) if CurrentContext() eq 'PML_En_T_SE';
  $this->{functor}=$_[0];
  after_edit_attr_hook($this,'functor',$_[0]);
}

sub FuncRSTR {
  AssignFunctor('RSTR');
}

sub FuncACT {
  AssignFunctor('ACT');
}

sub FuncPAT {
  AssignFunctor('PAT');
}

sub FuncPRED {
  AssignFunctor('PRED');
}

sub FuncCONJ {
  AssignFunctor('CONJ');
}

sub FuncLOC {
  AssignFunctor('LOC');
}

sub FuncTWHEN {
  AssignFunctor('TWHEN');
}

sub FuncRHEM {
  AssignFunctor('RHEM');
}

sub FuncDENOM {
  AssignFunctor('DENOM');
}

sub FuncMANN {
  AssignFunctor('MANN');
}

sub FuncEXT {
  AssignFunctor('EXT');
}

sub FuncID {
  AssignFunctor('ID');
}

sub FuncMAT {
  AssignFunctor('MAT');
}

sub FuncDIR3 {
  AssignFunctor('DIR3');
}

sub FuncDIR1 {
  AssignFunctor('DIR1');
}

sub FuncBEN {
  AssignFunctor('BEN');
}

sub FuncFPHR {
  AssignFunctor('FPHR');
}

sub FuncPREC {
  AssignFunctor('PREC');
}

sub FuncCRIT {
  AssignFunctor('CRIT');
}

sub FuncCAUS {
  AssignFunctor('CAUS');
}

sub FuncREG {
  AssignFunctor('REG');
}

sub FuncTHL {
  AssignFunctor('THL');
}

sub FuncADVS {
  AssignFunctor('ADVS');
}

sub FuncTHO {
  AssignFunctor('THO');
}

sub FuncCOMPL {
  AssignFunctor('COMPL');
}

sub FuncCOND {
  AssignFunctor('COND');
}

sub FuncOPER {
  AssignFunctor('OPER');
}

sub FuncAUTH {
  AssignFunctor('AUTH');
}

sub FuncCSQ {
  AssignFunctor('CSQ');
}

sub FuncSUBS {
  AssignFunctor('SUBS');
}

sub FuncTFRWH {
  AssignFunctor('TFRWH');
}

sub FuncTOWH {
  AssignFunctor('TOWH');
}

sub FuncDIR2 {
  AssignFunctor('DIR2');
}

sub FuncVOCAT {
  AssignFunctor('VOCAT');
}

sub FuncAPP {
  AssignFunctor('APP');
}

sub FuncADDR {
  AssignFunctor('ADDR');
}

sub FuncEFF {
  AssignFunctor('EFF');
}

sub FuncPAR {
  AssignFunctor('PAR');
}

sub FuncAPPS {
  AssignFunctor('APPS');
}

sub FuncACMP {
  AssignFunctor('ACMP');
}

sub FuncAIM {
  AssignFunctor('AIM');
}

sub FuncMEANS {
  AssignFunctor('MEANS');
}

sub FuncCM {
  AssignFunctor('CM');
}

sub FuncDIFF {
  AssignFunctor('DIFF');
}

sub FuncCPHR {
  AssignFunctor('CPHR');
}

sub FuncCPR {
  AssignFunctor('CPR');
}

sub FuncTSIN {
  AssignFunctor('TSIN');
}

sub FuncTPAR {
  AssignFunctor('TPAR');
}

sub FuncORIG {
  AssignFunctor('ORIG');
}

sub FuncMOD {
  AssignFunctor('MOD');
}

sub FuncATT {
  AssignFunctor('ATT');
}

sub FuncCNCS {
  AssignFunctor('CNCS');
}

sub FuncCONFR {
  AssignFunctor('CONFR');
}

sub FuncCONTRA {
  AssignFunctor('CONTRA');
}

sub FuncCONTRD {
  AssignFunctor('CONTRD');
}

sub FuncDISJ {
  AssignFunctor('DISJ');
}

sub FuncDPHR {
  AssignFunctor('DPHR');
}

sub FuncGRAD {
  AssignFunctor('GRAD');
}

sub FuncHER {
  AssignFunctor('HER');
}

sub FuncINTF {
  AssignFunctor('INTF');
}

sub FuncINTT {
  AssignFunctor('INTT');
}

sub FuncPARTL {
  AssignFunctor('PARTL');
}

sub FuncREAS {
  AssignFunctor('REAS');
}

sub FuncRESL {
  AssignFunctor('RESL');
}

sub FuncRESTR {
  AssignFunctor('RESTR');
}

sub FuncTFHL {
  AssignFunctor('TFHL');
}

sub FuncTTILL {
  AssignFunctor('TTILL');
}

sub FuncDESCR {
  AssignFunctor('DESCR');
}

sub FuncSM {
  AssignFunctor('SM');
}

sub FuncNE {
	AssignFunctor('NE');
}

1;

=back

=cut

