#-*- cperl -*-

use strict;

BEGIN
{
	import PML_T;
}

#include common.inc

#sub find_p_from_t_node
#{
#	my $t_node = shift;
#	return unless $t_node;
#
#	my $a_node = PML_T::GetANodeByID($t_node->attr('a/lex.rf'));
#	return find_p_from_a_node($a_node);
#}

#sub find_p_from_a_node
#{
#	my $a_node = shift;
#	return unless $a_node;
#	my ($p_node) = grep { defined $_->{tag} } PML_En_A::GetPNodes($a_node);
#	
#	return $p_node;
#}

sub coord001
{
	my $node = shift;
	return ('coord001') if $node->{is_member} and not (IsCoord($node->parent) or $node->parent->{functor} eq 'NE');
	return;
}

sub coord002
{
	my $node = shift;
	return ('coord002') if IsCoord($node) and scalar(grep { $_->{is_member}} $node->children) < 2;
	return;
}

sub coord003
{
	my $node = shift;
    return ('coord003') if $node->{is_member} && $node->{functor} =~ m/^(?:CM|PREC|FPHR)$/;
	return;
}

## 1. VOCAT and PARTL can be coordinated only within themselves.
## 2. PRED and DENOM can be coordinated only within themselves.
## 3. PAR can be coordinated only with PAR.
## 4. Actant can be coordinated only with an actant of the same type.
sub coord004
{
	my $node = shift;
	my @output;
  	my $functor;
	if ($node->{functor} =~ m/^(?:VOCAT|PAR(?:TL)?|PRED|DENOM|ACT|PAT|ADDR|EFF|ORIG)$/ and $node->{is_member})
	{
    	$functor = $node->{functor};
	    my $head=$node;
    	while($head and $head->{is_member})
		{
			$head=$head->parent;
		}
    	my @expand = grep $_->{is_member}, $head->children;
    	while(first {$_->{is_member} and IsCoord($_)} @expand)
		{
      		@expand = grep $_,map { $_->{is_member} ? (IsCoord($_) ? grep $_->{is_member},$_->children : $_) : undef; } @expand;
    	}
    	if ($functor =~ m/^(?:VOCAT|PARTL)$/ and first { $_->{functor} !~ m/^(?:VOCAT|PARTL)$/ } @expand)
		{
      		push @output, 'coord004.1';
    	}
		elsif ($functor =~ m/^(?:PRED|DENOM)$/ and first { $_->{functor} !~ m/^(?:PRED|DENOM)$/ } @expand)
		{
      		push @output, 'coord004.2';
    	}
		elsif ($functor eq 'PAR' and first { $_->{functor} ne 'PAR' } @expand)
		{
      		push @output, 'coord004.3';
    	}
		elsif ($functor =~ m/(?:ACT|PAT|ADDR|EFF|ORIG)/ and first{ $_->{functor} ne $functor } @expand)
		{
      		push @output, 'coord004.4';
    	}
  	}

	return @output;
}

sub coord005
{
	my $node = shift;
  if($node->{functor}eq'OPER'){
    my @expand=grep $_->{is_member},$node->children;
    while(first {$_->{is_member} and IsCoord($_)} @expand){
      @expand = grep $_,map {
        $_->{is_member}
          ? (IsCoord($_) ? grep $_->{is_member},$_->children : $_)
            : undef;
      } @expand;
    }
    my %func;
    @func{map {$_->{functor}} @expand}=();
    return ('coord005') if keys(%func) > 1;
  }

  return;
}

## 1. RHEM on a coordination head is never a member.
## 2. Coordination head with a RHEM child must have a common dependant
##   child as well with functor different form RHEM and CM.
sub coord006
{
	my $node = shift;
	my @output;
  if(IsCoord($node)){
    my @ch = $node->children;
    my @rhem = grep $_->{functor} eq 'RHEM',@ch;
    if(@rhem){
      push @output, 'coord006.1' if first { $_->{is_member} } @rhem;
      push @output, 'coord006.2' if not first {
          ! $_->{is_member} and $_->{functor} !~ m/^(?:RHE|C)M$/
        } @ch;
    }
  }

  return @output;
}

## 1. Every a-node has a t-node and m-node
## 2. M_a={t,a in t->lex}. M_a != 0 -> exists exactly one t in M not
##    generated
## 3. Nothing can be both in lex and aux.
## 4. All nodes with non-# lemma have lex node.
## 5. Some # lemmas (punctutation) have only limited set of lexes or
##    auxes.
## 6. If something is in aux of several nodes, all of them are
##    coordinated in one coordination, or they have the same t_lemma,
##    or the t-lemmas of their parents are the same, or one of the
##    nodes is a root with a PREC child.

sub coord_head {
  my $node = $_[0];
  while ($node->{is_member} && $node->parent){
    $node = $node->parent;
  }
  return $node;
} #coord_head

sub coref_gram001 
{
	my $node = shift;
	
	return if $node->{t_lemma} ne "#Cor";

	return ('coref_gram001.1') if not ($node->{'coref_gram.rf'});
 	
	if($node->{'coref_gram.rf'}){
	   my @ref = ListV($node->{'coref_gram.rf'});
    	my @target;
    	foreach my $ref (@ref) {
      		push @target,$ref if GetNodeByID($ref);
    	}
    	return 'coref_gram001.2' if scalar(@ref) != scalar(@target);
  	}

	return;
}

sub bracket001{
	my $node = shift;

	my $a_node = PML_T::GetALexNode($node);
	if (not defined $a_node and $node->{t_lemma} eq "#Bracket"){
		return 'bracket001.3';
	}

	if ($node->{t_lemma} eq "#Bracket" and lc($a_node->{'m'}->{'form'}) !~ "^(-|)(lrb|rrb|lcb|rcb)(-|)\$"){
		return 'bracket001.1';
	}
	
	if ($node->{t_lemma} ne "#Bracket" and lc($a_node->{'m'}->{'form'}) =~ "^(-|)(lrb|rrb|lcb|rcb)(-|)\$"){
		return 'bracket001.2';
	}
	
	return;

}

sub links001 {
	my ($engvallex, $prefix, @trees) = @_;
	my @output;
  my %anode;
  my $node;
  my %punct = ('#Comma' => ',',
               '#Colon' => ':',
               '#Dash' => '-',
 #             '#Dash' => '--',
               '#Percnt' => '%',
               '#Period' => '.',
               '#Slash' => '/',
               '#Semicolon' => ';',
               '#Amp' => '&');
  my %a_punct = (',' => ',',
               ':' => ':',
               '-' => '-',
			   '--' => '--',
               '%' => '%',
               '.' => '.',
               '/' => '/',
               ';' => ';',
               '&' => '&');

# hash part

	foreach my $tree (@trees)
	{
	$node = $tree;
    while ($node) {
      if($node->attr('a/lex.rf')){
        push @{$anode{$node->attr('a/lex.rf')}{lex}},$node;
      }
      if (my $auxl = $node->attr('a/aux.rf')) {
        foreach my $aux (ListV($auxl)) {
          push @{$anode{$aux}{aux}},$node;
        }
      }
    } continue {
      $node = $node->following;
    }
	}

# test 2 & 3

  foreach my $lex (grep $anode{$_}{lex},keys %anode) {
    my @nongen = grep ! $_->{is_generated},@{$anode{$lex}{lex}};
    my $segmentationError = first {
      first {$_->{type} eq 'Word segmentation'} ListV($_->{annot_comment});
    } @nongen;
	my $a_node = PML_T::GetANodeByID($lex);
	my $is_frac = ($a_node->attr('m/form') =~ m/^[0-9]+\\\/[0-9]+$/ ? 1 : 0);
    if (@nongen > 1 and not $is_frac and $a_node->attr('m/tag') ne '-NONE-'){
      if ($segmentationError){
        push @output, "links001.2.1-warning " . join(" ", map { $_->{id} } @{$anode{$lex}{lex}});
      }else{
        push @output, "links001.2 " . join(" ", map { $_->{id} } @{$anode{$lex}{lex}});
      }
    }
    if($anode{$lex}{aux}){
      my @nongen = grep ! ( $_->{is_generated} or $_->{t_lemma} =~ /^(be|must|have|can|might|may)$/ )
        ,@{$anode{$lex}{lex}},@{$anode{$lex}{aux}} ;
      my $segmentationError = first {
        first {$_->{type} eq 'Word segmentation'} ListV($_->{annot_comment});
      } @nongen;
      if (@nongen > 1){
        if ($segmentationError){
            push @output, "links001.3.1-warning " . join(" ", map { $_->{id} } @{$anode{$lex}{aux}},@{$anode{$lex}{lex}});
        }else{
            push @output, "links001.3 " . join(" ", map { $_->{id} } @{$anode{$lex}{aux}},@{$anode{$lex}{lex}});
        }
      }
    }
  }

# test 6

  foreach my $aux (grep $anode{$_}{aux},keys %anode){
	my $form = (PML_T::GetANodeByID($aux))->attr('m/form');
	next if $form =~ m/^[,.?!;:]$/;

    my (%tlemma,%parent);
    @tlemma{ map { $_->{t_lemma} } @{$anode{$aux}{aux}} } = ();
    @parent{ map {
      $_->{is_member}
        ? coord_head($_)
          : (IsCoord($_->parent) ? $_ : $_->parent)->{t_lemma}
        } @{$anode{$aux}{aux}} } = ();
    if(keys %tlemma > 1 and keys %parent > 1){
        # previous sentence
        unless (uniq(map { $_->root } @{$anode{$aux}{aux}}) > 1
                and first {
                  $_->parent == $_->root
                    and first { $_->{functor} eq 'PREC' } $_->children
                  } @{$anode{$aux}{aux}}
               ){
			   push @output, "links001.6.1 " . join(" ", map { $_->{id} } @{$anode{$aux}{aux}});
      }
    }
  }

  foreach my $tree (@trees)
  {
  $node = $tree;
    while ($node) {

# test 4

      if($node->parent and $node->{t_lemma} !~ m/^#/){
        if (! $node->attr('a/lex.rf')){
          push @output, "links001.4 $node->{id}";
        }
      }

# test 5

      if(exists $punct{$node->{t_lemma}}){
        push @output, "links001.5.1 $node->{id}" if $node->{is_generated};
        if ($node->attr('a/lex.rf')){
          push @output, "links001.5.3 $node->{id}" unless
            (PML_T::GetANodeByID($node->attr('a/lex.rf'))->attr('m/form') eq $punct{$node->{t_lemma}}
			 or ($node->{t_lemma} eq '#Dash' and PML_T::GetANodeByID($node->attr('a/lex.rf'))->attr('m/form') eq '--' )	
			 or ( $node->{t_lemma} eq '#Slash' and PML_T::GetANodeByID($node->attr('a/lex.rf'))->attr('m/form') =~ m/^[0-9]+\\\/[0-9]+$/ )	
             or # "2.-3."
             $node->{t_lemma} eq '#Dash'
             && PML_T::GetANodeByID($node->attr('a/lex.rf'))->attr('m/form') =~ /^[0-9]*-[0-9]+$/
             && first { $_->{type} eq 'Word segmentation' } ListV($node->{annot_comment}));
        }else{
          push @output, "links001.5.2 $node->{id}";
        }
        push @output, "links001.5.4-warning $node->{id}" if ($node->{t_lemma} ne '#Percnt') && (scalar(ListV($node->attr('a/aux.rf')))) and grep { not exists $a_punct{$_->attr('m/lemma')} } map { PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf'));
      }elsif($node->{t_lemma} eq '#Period3'){
        push @output, "links001.5.5 $node->{id}" unless $node->attr('a/lex.rf') and join('',map {$_->attr('m/form')} PML_T::GetANodes($node)) eq '...';
      }
      if($node->{t_lemma} eq '#Separ'){
        push @output, "links001.5.6 $node->{id}" if ! $node->{is_generated} or defined $node->{a}{'lex.rf'};
      }
    } continue {
      $node = $node->following;
    }
	}

# test 1
    foreach my $tree (@trees)
	{
	my $atree = PML_T::GetANodeByID($tree->{'atree.rf'});
    my $anode = $atree;
    while($anode = $anode->following){
      if($anode->attr('m/lemma') =~ m/^(?:[[:alnum:]])/ and $anode->attr('m/tag') ne '-NONE-' and
         ! exists $anode{'a#'.$anode->{id}}
         and ! first {
           my $mf = $anode->attr('m/form');
           my $ml = $anode->attr('m/lemma');
           $ml =~ s/[-_'`].+//;
           $_->{type} eq 'Translation'
             and $_->{text} =~ m/\b\Q$mf\E\b|\b\Q$ml\E\b/i
           } ListV($tree->{annot_comment})){
		   push @output, "links001.1.1-" . $anode->attr('m/form') . "-" . $anode->{id} . " " . $tree->{id};
      }
    }
	}

  return @output;
}#links001

sub links002
{
	my $node = shift;
	my @output;

  my @ref;
  my @target;
  push @ref,$node->attr('a/lex.rf') if $node->attr('a/lex.rf');
  push @ref,$node->attr('atree.rf') if $node->attr('atree.rf');
  push @ref,ListV($node->attr('a/aux.rf')) if $node->attr('a/aux.rf');
  for my $ref (@ref)
  {
	  my $a = PML_T::GetANodeByID($ref);
	  push @target, $ref if defined $a->{id};
  }
  my $ref = scalar @ref;
  my $target = scalar @target;
  push @output, "links002.1-$ref:$target" if $ref != $target;

# compl.rf

  if($node->{'compl.rf'}){
    @ref = ListV($node->{'compl.rf'});
    my @target;
    foreach my $ref (@ref) {
      push @target,$ref if GetNodeByID($ref);
    }
    push @output, 'links002.2' if scalar(@ref) != scalar(@target);
  }

  return @output;
}

sub links003
{
	my $node = shift;
	return ('links003') if $node->{functor} =~ m/^(?:DENOM|VOCAT)$/ and $node->attr('a/aux.rf') and
	                    grep { $_->attr('m/lemma') !~ m/^(a|an|the)$/ and $_->attr('m/tag') !~ m/^(?:[.,:]|''|``|-RRB-|-NONE-)$/ } map { PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf'));
	return;
}

sub structure001
{
	my $node = shift;
	my @output;
  push @output, 'structure001.1' if $node->{functor}=~ m/^(?:PRED|DENOM)$/ and first {$_->{nodetype} ne 'root'} PML_T::GetEParents($node);
  push @output, 'structure001.2'
    if $node->{nodetype} eq 'root'
      and ((first {
        $_->{functor} !~ m/^(?:PRED|DENOM|VOCAT|PARTL|PAR|CONJ|DISJ|ADVS|REAS|GRAD|CONFR|CONTRA|APPS|CSQ|OPER)$/
      } $node->children) or first {
        $_->{functor} !~ m/^(?:PRED|DENOM|PAR|PARTL|VOCAT)$/
      } PML_T::GetEChildren($node));

  return @output;
}

sub is_nominativ
{
	my $node = shift;
	my ($tag_lex) = map { defined $_ ? $_->attr('m/tag') : '' } (PML_T::GetANodeByID($node->attr('a/lex.rf')));
	my ($form_lex) = map { defined $_ ? $_->attr('m/form') : '' } (PML_T::GetANodeByID($node->attr('a/lex.rf')));
	my @tag_aux = map { $_->attr('m/tag') } map { PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf'));
	return 1 if ($tag_lex eq 'CD' and $form_lex !~ m/(rd|th|st|nd)$/i) or
	           ($tag_lex =~ m/^NN/ and not grep { $_ =~ m/^(POS|IN)$/ } @tag_aux) or
			   ($tag_lex eq 'PRP' and $form_lex !~ m/^(me|him|her|us|them)$/i);
	return 0;
}

sub structure002
{
	my $node = shift;
	my @output;
  if ($node->{functor} eq 'PAR'){
    push @output, 'structure002.1' unless $node->{is_parenthesis};

    return if first { $_->{type} =~ /^M-/ } ListV($node->{annot_comment});

	my $OK = 0;
	$OK = 1 if $node->{t_lemma} =~ m/^(#Forn|#Idph|#EmpNoun|#EmpVerb)$/;
	my $a_node = PML_T::GetANodeByID($node->attr('a/lex.rf'));
	my $tag_lex = $a_node ? $a_node->attr('m/tag') : '';
	$OK = 1 if is_nominativ($node) or $tag_lex =~ m/^V/;
	$OK = 1 if $node->{t_lemma} =~ m/\$/;
	$OK = 1 if ($tag_lex eq "\$");
	push @output, 'structure002.2' if not $OK;
  }

  return @output;
}

sub structure003
{
	my $node = shift;
	my @output;
  if($node->{functor} eq 'FPHR'){
    push @output, 'structure003.4' if scalar(ListV($node->attr('a/aux.rf'))) > 0;
    push @output, 'structure003.2' if $node->parent and $node->parent->{t_lemma} ne '#Forn';
    if(not $node->attr('a/lex.rf')){
      push @output, 'structure003.1.1';
    }elsif( ( ($node->{functor} ne 'NE' and $node->{t_lemma} ne PML_T::GetANodeByID($node->attr('a/lex.rf'))->attr('m/form'))
		      or
              ($node->{functor} eq 'NE' and lc($node->{t_lemma}) ne lc(PML_T::GetANodeByID($node->attr('a/lex.rf'))->attr('m/form'))) )
           and not first {
             $_->{type} =~ /Word segmentation|Typo/
           } ListV($node->{annot_comment})){
      push @output, 'structure003.1';
    }
  }elsif($node->{t_lemma} eq '#Forn'){
    push @output, 'structure003.3' unless first {$_->{functor} eq 'FPHR' } $node->children;
  }

  return @output;
}

## 1. DPHR has no actants nor DPHR children
## 2. DPHR has no DPHR sisters
## 3. DPHR has empty val_frame.rf
## 4. DPHR has no New valency frame comment - not applicable
sub structure004 {
	my $node = shift;
	my @output;
  if($node->{functor} eq 'DPHR'){
    push @output, 'structure004.1' if first { $_->{functor} =~ m/^(?:ACT|PAT|ADDR|EFF|ORIG|DPHR)$/ } PML_T::GetEChildren($node);
    push @output, 'structure004.2' if first {$_ != $node and $_->{functor} eq 'DPHR'} map {PML_T::GetEChildren($_)} PML_T::GetEParents($node);
    push @output, 'structure004.3' if $node->{'val_frame.rf'};
  }

  return @output;
}

## 1. functor eq 'COMPL' iff compl.rf
## 2. COMPL is not adverb
## 3. nor prepositional phrase, except with "as
sub structure005 {
	my $node = shift;
	my @output;
  if (not ($node->{functor} ne 'COMPL' xor scalar(ListV($node->{'compl.rf'})) > 0)){
    push @output, 'structure005.1';
  }
  return if $node->{functor} ne 'COMPL';
	my $anode = PML_T::GetANodeByID($node->attr('a/lex.rf'));
	my $pnode = find_p_from_a_node($anode) if $anode;
	return if not $pnode;

	push @output, 'structure005.2' if $anode->attr('m/tag') =~ m/^RB/;
	while ($pnode->{is_head} and $pnode->parent) { $pnode = $pnode->parent }
	$pnode = $pnode->parent if $pnode->parent;
	return if $pnode->{phrase} ne 'PP';

	my $in_rb = first { $_->{tag} =~ m/^(IN|RB)$/ } $pnode->children;
	push @output, 'structure005.3' if $in_rb->{form} !~ m/^as$/i;

	return @output;
}

sub structure006
{
	my $node = shift;
  if ($node->{t_lemma} =~ m/#(?:Unsp|Gen)/
      and $node->{functor} !~ m/^(?:ACT|PAT|ADDR|EFF|ORIG|CPHR|DPHR)$/){
    if(not ($node->{functor} eq 'APP' and $node->{t_lemma} eq '#NewNode')){
      return ('structure006');
    }
  }
  return;
}

sub structure007
{
	my $node = shift;
  if ($node->{functor} eq 'ID'){
    my @p = GetEParents($node);
	my $OK = 0;
	for my $p (@p)
	{
		my $a_p = PML_T::GetANodeByID($p->attr('a/lex.rf'));
		$OK = 1 if $p->{t_lemma} eq '#Idph';
		$OK = 1 if $a_p and $a_p->attr('m/tag') =~ m/^N/;
	}
	return ('structure007') if not $OK;
    }
	return;
}

sub structure008
{
	my $node = shift;
	return if $node->{functor} ne 'CPHR';
	return ('structure008') if not $node->parent;
	return ('structure008') if not first { $_->{functor} eq "PRED" or $_->{t_lemma} eq "#EmpVerb" or get_p_tag($_) =~ m/^V/ } PML_T::GetEParents($node);
	return;
}

sub structure009
{
	my $node = shift;
	return ('structure009') if $node->{functor} eq 'CM' and (not $node->parent or not IsCoord($node->parent));
	return;
}

sub structure010
{
	my $node = shift;
	return ('structure010') if $node->{functor} =~ m/^(DENOM|VOCAT)$/ and not (is_nominativ($node) or $node->{t_lemma} =~ m/^(?:#EmpNoun|#Forn|#Idph|\$|#)$/);
	return;
}

sub structure014
{
	my $node = shift;
	my %ch = (AsMuch=>'RESL',
    	      Equal=>'CPR',
        	  Total=>'RESTR');
my $lemmas = join '|',keys %ch;

  if($node->{t_lemma} =~ /^#(?:$lemmas)$/){
    my @ch = GetEChildren($node);
    unless(@ch){
      return ('structure014.1');
    }else{
      return ('structure014.2')
        if first {
          $_->{functor} ne $ch{substr($node->{t_lemma},1)}
        } @ch;
    }
  }
  return;
}

sub structure015
{
	my $node = shift;
    return ('structure015') if $node->{t_lemma} eq '#Some' and $node->{functor} ne 'PAT';
	return;
}

sub structure017
{
	my $node = shift;
  if($node->{t_lemma} eq '#Oblfm'
     and $node->{functor} =~ m/^(?:ACT|ADDR|EFF|ORIG)$/){
    return ('structure017-warning');
  }
  return;
}

sub structure018
{
	my $node = shift;
	my @output;
  if($node->{t_lemma} eq '#Bracket'){
    push @output, 'structure018.1' if $node->{functor} ne 'APPS';
    push @output, 'structure018.2' unless $node->{is_parenthesis};
  }
  return @output;
}

sub structure019
{
	my $node = shift;
	return ('structure019') if $node->{t_lemma} =~ m/^#(Comma|Colon|Dash|Period|Slash|Semicolon)$/ and $node->{functor} !~ m/^(PRED|PAR|CONJ|DISJ|ADVS|REAS|GRAD|CONFR|CONTRA|APPS|CSQ|OPER|PAT)$/;
	return;
}

sub structure020
{
	my $node = shift;
  if($node->{t_lemma} =~ /^#Idph$/){
    my @ch = GetEChildren($node);
    unless(@ch){
      return ('structure020.1');
    }else{
      return ('structure020.2')
        if not first {
          $_->{functor} eq 'ID'
        } @ch;
    }
  }
  return;
}

sub structure021
{
	my $node = shift;
	return ('structure021') if $node->{t_lemma} =~ m/^#Emp(Noun|Verb)$/ and scalar(ListV($node->attr('a/aux.rf'))) == 0 and not $node->children;
	return;
}

sub structure022
{
	my $node = shift;
	return unless $node->{is_generated} and $node->{t_lemma} !~ m/^#/ and scalar($node->children) == 0;
	return if $node->{functor} =~ m/^(CPHR|DPHR)$/;
  
	my $second;
	my $r = $node;
	while ($r->parent) { $r = $r->parent }

	foreach my $n ($r->descendants) {
		if ($n->attr('a/lex.rf') eq $node->attr('a/lex.rf') and not $n->{is_generated}) {
			$second = $n;
			last;
		}
	}
	if ($second) {
		my @aux1 = grep {
		             $_->attr('m/lemma') =~ m/^(?:be|have|could|should|would|can|may|might|will|shall|ought|must)$/
								 or $_->attr('m/tag') =~ m/^(?:IN|RP)$/
							 } map { PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf'));
		my @aux2 = grep {
		             $_->attr('m/lemma') =~ m/^(?:be|have|could|should|would|can|may|might|will|shall|ought|must)$/
								 or $_->attr('m/tag') =~ m/^(?:IN|RP)$/
							 } map { PML_T::GetANodeByID($_) } ListV($second->attr('a/aux.rf'));
		my @intersect = TredMacro::ListIntersect(\@aux1, \@aux2);
		return if scalar(@intersect) != scalar(@aux1) or scalar(@intersect) != scalar(@aux2);

		my $neg1 = (grep { $_->{t_lemma} eq '#Neg' } $node->parent->children) ? 1 : 0;
		my $neg2 = (grep { $_->{t_lemma} eq '#Neg' } $second->parent->children) ? 1 : 0;
		return if $neg1 != $neg2;
	  
		my ($t1, $t2) = map { $_->attr('m/tag') } map { PML_T::GetANodeByID($_->attr('a/lex.rf')) } ($node, $second);
		return if (($t1 =~ m/^JJ/ and $t2 =~ m/^JJ/) or ($t1 =~ m/^RB/ and $t2 =~ m/^RB/)) and $t1 ne $t2;
	}
	
	if ( first { $_->{functor} eq 'CPR' } PML_T::GetEParents($node) ){
		return if get_p_tag($node) =~ m/^(?:JJ|RB)/ ;

		if (get_p_tag($node) =~ m/^NN/){
			my @aux = grep { $_->attr('m/tag') eq "IN" } map { PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf'));
			return if scalar(@aux);
		}

	}	

	return if $node->{functor} eq 'PAT' and get_p_tag($node) =~ m/^(?:JJ|RB)/
	          and first { $_->{t_lemma} =~ m/^(?:be|get|feel|look|sound|go|drive|make)$/ } PML_T::GetEParents($node);
	
	return ('structure022');
}

sub structure023
{
	my $node = shift;
	return if $node->{nodetype} ne 'root';
  my $num = first {$_->{type} eq 'Sentence segmentation'} ListV($node->{annot_comment});
  $num = $num->{text} if $num;
  $num =~ s/ .*//;
  if($node->children != 1 && $node->children != $num){
    return ('structure023');
  }
  return;
}

sub structure026
{
	my $node = shift;
	return ('structure026') if $node->{functor} eq '???' or $node->{functor} eq '---';
	return ('structure026') if $node->{functor} =~ m/\|/;
	return;
}

sub get_p_tag
{
	my $t_node = shift;
	my $a_node = PML_T::GetANodeByID($t_node->attr('a/lex.rf'));
	return '' if not $a_node;
	return $a_node->attr('m/tag');
}

sub structure027
{
	my $node = shift;
	if ($node->{functor} eq 'PREC')
	{
		my $check = 0;
		$check = 1 if $node->parent->{functor} =~ m/^(?:PRED|PAR|DENOM|PARTL)$/;
		$check = 1 if IsCoord($node->parent) and not first { $_->{functor} !~ m/^(?:PRED|PAR|PRED|PARTL)$/ } grep { $_->{is_member} } PML_T::GetEChildren($node->parent);
		if (not $check and $node->parent and $node->parent->parent)
		{
			my $parent = $node->parent;
			if (get_p_tag($parent) =~ m/^V/ or (IsCoord($parent) and grep {/^V/} map { get_p_tag($_) } $parent->children))
			{
				$parent = $parent->parent;
				while (get_p_tag($parent) !~ m/^V/ and $parent->parent) { $parent = $parent->parent; }
				if (get_p_tag($parent) =~ m/^V/ and $parent->{t_lemma} =~ m/^(?:say|add|claim|report|tell|conclude)$/) { $check = 1; }
			}
			
		}
		if (not $check and $node->parent->{t_lemma} eq '#EmpVerb')
		{
			#This allows various list enumerating, e.g. 22a), XII:, A. etc.
			$check = 1 if $node->{t_lemma} =~ m/^(([ivxlcdm]+)|([0-9]+)){0,1}[a-z]{0,2}[^a-z]{0,1}$/i;
		}
		return ('structure027') if not $check;
	}
	return;
}

sub structure028
{
	my $node = shift;
	return unless ($node->{functor} eq 'PRED' and $node->{t_lemma} !~ /_/);
	my @aux_tags = map { $_->attr('m/tag') } map { PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf'));
	return ('structure028') if grep { /^IN$/ } @aux_tags;
	return;
}

sub structure029
{
	my $node = shift;
	return ('structure029') if $node->{t_lemma} eq '#Neg' and $node->{is_generated};
	return;
}

sub check_p_node_s030
{
	my $p_node = shift;
	return unless $p_node->{phrase} =~ m/^(NP|NML)$/;

	my ($nac_loc) = grep { $_->{phrase} eq 'NAC' and grep { /LOC/ } ListV($_->{functions}) } $p_node->children;
	return if not $nac_loc;

	my $np = $nac_loc->rbrother;
	while ($np and $np->{phrase} !~ m/^(NP|NML)$/ and $np->{tag} !~ m/^NN/) { $np = $np->rbrother }
	return if not $np;

	my @ch = $nac_loc->children;
	return if scalar(@ch) < 3 or scalar(@ch) > 4;
	return unless $ch[1]->{tag} =~ m/^(,|-LRB-)$/;
	return if $ch[3] and $ch[3]->{tag} !~ m/^(,|-RRB-)$/;

	my $third_ok = 0;
	$third_ok = 1 if $ch[2]->{phrase} =~ m/^(NP|NML)$/ and not first { $_ ne 'GPE:STATE_PROVINCE' } map { $_->{bbn_tag} } grep { defined $_->{tag} } $ch[2]->descendants;
	$third_ok = 1 if defined $ch[2]->{tag} and $ch[2]->{bbn_tag} eq 'GPE:STATE_PROVINCE';
	$third_ok = 1 if $ch[2]->{form} =~ m/^[IVXLCDM]+\.?$/ and not $ch[2]->{bbn_tag};
	return if not $third_ok;

	if (not $ch[3])
	{
		my $right = $nac_loc->rbrother;
		while (not defined $right->{tag}) { $right = $right->firstson }
		$ch[3] = $right if $right->{tag} =~ m/^(,|-RRB-)$/;
	}

	return (find_head($ch[0]), $ch[1], find_head($ch[2]), $ch[3]);
}

sub structure030
{
	my $t_tree = shift;
	my @output = ();

	my ($a_tree) = PML_T::GetANodes($t_tree->root);
	my ($p_tree) = PML_En_A::GetPNodes($a_tree->root);
	
	foreach my $p_node ($p_tree, $p_tree->descendants)
	{
		my ($one, $two, $three, $four) = check_p_node_s030($p_node);
		next unless $one;

		my $t_three = find_t_node($three->{id}, $a_tree, $root);
		next unless $t_three;
		push @output, "structure030 $t_three->{id}" if $t_three->{functor} ne 'PAR';
	}

	return @output;
}

sub structure031
{
	my $t_tree = shift;
	my @output = ();
	
	foreach my $node ($t_tree->descendants)
	{
		my @members = grep { $_->{is_member} } $node->children;
		my %nonmemb_func;
		foreach my $func (map { $_->{functor} } grep { not $_->{is_member} } $node->children)
		{
			$nonmemb_func{$func} = ();
		}
		foreach my $child (map { $_->children } @members)
		{
			push @output, "structure031 $child->{id}" if $child->{functor} =~ m/^(?:ACT|PAT|ADDR|EFF|ORIG)$/ and exists $nonmemb_func{$child->{functor}};
		}
	}
	
	return @output;
}

sub structure032
{
	my $node = shift;

	return if ($node->{t_lemma} =~ /_/);
	return if ($node->{is_generated} and $node->{t_lemma} !~ /^#/ );
	my $a_node = PML_T::GetANodeByID($node->attr('a/lex.rf')) if $node->attr('a/lex.rf');
	return unless $a_node;
	
	return if ( grep { $_->{type} =~ m/^Wrong m-tag/ } ListV($node->{annot_comment}));
	my $a_be = 0;
	my $a_it = 0;
	my $struct032 = 0;
	foreach my $aux (ListV($node->attr('a/aux.rf')))
	{
		my $a_aux = PML_T::GetANodeByID($aux);
		$a_it = 1 if $a_aux->attr('m/lemma') eq "it";
		$a_be = 1 if $a_aux->attr('m/lemma') eq "be";
		next if $a_aux->attr('m/tag') ne 'IN' or $a_aux->attr('m/form')=~/^(ago|like|as|than|notwithstanding)$/;
		$struct032 = 1 if $a_aux->{ord} == ($a_node->{ord} + 1);
	}
	
	return ('structure032-warning') if $struct032 and (not $a_it or not $a_be);
	
	return;
}

sub structure033
{
	my $node = shift;
	my $a_node = PML_T::GetANodeByID($node->attr('a/lex.rf')) if $node->attr('a/lex.rf');
	
	if ($a_node and $a_node->attr('m/tag') =~ m/^(?:NNS|NNPS)$/){
		my $additional_lemma = 0;
		foreach my $child (PML_T::GetEChildren){
			if ($child->{t_lemma} eq "additional"){
				$additional_lemma = 1;
				last;
			};
		}

		if (not $additional_lemma){
			foreach my $aux (ListV($node->attr('a/aux.rf')))
			{
				my $a_aux = PML_T::GetANodeByID($aux);
				return ('structure033.1') if $a_aux->attr('m/lemma') =~ m/^a(|n)$/;
			}
		}
	}

	if ($node->{is_member} and $a_node and $a_node->attr('m/tag') =~ m/^(?:NN|NNS)$/ and not grep { $_->attr('m/tag') eq 'DT' } map {PML_T::GetANodeByID($_) } ListV($node->attr('a/aux.rf')))
	{
		my @ch = PML_T::ExpandCoord(PML_T::GetNearestNonMember($node));
		@ch = sort { $a->{deepord} <=> $b->{deepord} } @ch;
		if ($ch[0]->{id} ne $node->{id})
		{
			my ($article) = grep { $_->attr('m/tag') eq 'DT' } map { PML_T::GetANodeByID($_) } ListV($ch[0]->attr('a/aux.rf'));
			if ($article and ($article eq "the" or $a_node->attr('m/tag') eq "NN"))
			{
				return ('structure033.2-warning');
			}
		}
	}

	return;
}

sub structure034
{
	my $node = shift;
	return unless $node->{bbn_tag} and $node->{functor} !~ /^(NE|CONJ)$/ and $node->parent and $node->parent->{t_lemma} ne '#Forn';
	
	my $par = $node->parent;
	while ($par->parent and ($par->{bbn_tag} eq $node->{bbn_tag}) and ($par->{functor} ne 'NE'))
	{
		$par = $par->parent;
	}
	return ('structure034') if $par->{functor} eq 'NE' and $par->{bbn_tag} eq $node->{bbn_tag};
	return;
}

sub structure035
{
	my ($node, $engvallex, $prefix) = @_;
	return unless $node->parent and $node->{t_lemma} eq '#NewNode';
	my $a_parent = PML_T::GetANodeByID($node->parent->attr('a/lex.rf'));
	return unless $a_parent;

	if (grep {$_->{type} =~ m/^Nominalisation/} ListV($node->{annot_comment})){
		return;
	}
	
	if ($a_parent->attr('m/tag') =~ m/^NN/){
		if ($node->{functor} eq 'APP'){
			return;
		}else {
			if ($node->{functor} ne 'ACT'){
				return ('structure035');
			}else{
				if ($node->parent->{t_lemma} =~ m/(age|al|ation|ison|ion|ing|ment|ant|ee|er|or)$/i){
					return;
				}else{
					return ('structure035');					
				
				}
		
			}
		}
	}
	return;

}

sub structure036
{
	my $node = shift;
	return unless $node->{t_lemma} eq 'about' and $node->{functor} eq 'EXT';
	return ('structure036-warning ') unless $node->parent;
	return if (get_m_lemma($node->parent) =~ m/^(?:more|less|fewer|equal)$/i);
	if (get_p_tag($node->parent) ne 'CD' or $node->parent->{functor} ne 'RSTR'){
		my $number_desc=0;
		my @descendants=$node->parent->descendants;
		foreach my $desc (@descendants){
			if($desc->{functor} eq 'RSTR' and get_p_tag($desc) eq 'CD'){
				$number_desc=1;
				last;
			}
		}
		return ('structure036-warning ') if ($node->parent->{nodetype} ne 'coap' and $number_desc);
	}
	return;
}

sub get_m_lemma
{
	my $t_node = shift;
	my $a_node = PML_T::GetANodeByID($t_node->attr('a/lex.rf'));
	return '' if not $a_node;
	return $a_node->attr('m/lemma');
}

sub structure037
{
	my $node = shift;
	return if $node->{nodetype} eq 'root';
	return ('structure037') if not $node->{t_lemma} or $node->{t_lemma} =~ m/ /;
	return;
}

sub t_lemma001
{
	my $node = shift;
	return ('t_lemma001-' . $node->{t_lemma}) if $node->{t_lemma} eq "#NewNode";
	return;
}

sub valency001
{
	my $node = shift;
	return if not $node->{a}{'lex.rf'};
	return if grep { $_->{type} =~ m/^(Engvallex|Wrong m-tag)/ } ListV($node->{annot_comment});
	#	return if $node->{t_lemma} =~ m/^(be|have|have_got)$/;
	return if $node->{functor} =~ m/^(D|F)PHR$/;
	my $anode = PML_T::GetANodeByID($node->attr('a/lex.rf'));
	return ('valency001')	if (
		                     (IsAlt($node->{'val_frame.rf'}) and scalar(AltV($node->{'val_frame.rf'})) != 1)
							 or
							 (not IsAlt($node->{'val_frame.rf'}) and $node->{'val_frame.rf'} eq '')
						   )
						   and
						   $anode->attr('m/tag') =~ m/^V/;
	return;
}

sub valency002 {
	my $node = shift;
    my @ch = grep $_->{functor} =~ m/^(?:ACT|PAT|ADDR|EFF|ORIG)$/, GetEChildren($node);
  if( scalar(@ch) ){
    my %func;
    @func{map { $_->{functor} } @ch}=();
    foreach my $func (keys %func){
      my @same = grep $_->{functor} eq $func,@ch;
      @same = map {
        until(! $_->{is_member} or ! $_->parent){
          $_=$_->parent
        }
        $_;
      } @same;
      my %same;
      @same{@same}=();
      if (keys(%same) > 1) {
        return ('valency002');
      }
    }
  }
  return;
}#valency

sub valency003 {
	my ($node, $engvallex, $prefix) = @_;
	my @output;
	#	my ($prefix,$file) = FindVallex();
	#	my $engvallex = TrEd::EngValLex::NonGUI->new($file);
  
	return unless
        ($node->{functor} eq 'CPHR'
         or ($node->attr('a/lex.rf')
             and (GetANodeByID($node->attr('a/lex.rf'))->attr('m/tag') =~ /^V/)));
	  return if grep { $_->{type} =~ m/^Engvallex/ } ListV($node->{annot_comment});

      my $rf = $node->{'val_frame.rf'};
      return unless $rf;
      $rf=@{$rf}[0] if ref $rf;
      $rf =~ s/^\Q$prefix\E\#//;

	  my $frame = $engvallex->by_id($rf);
	  
	  return ("valency003.4") if not $frame;
	  return ("valency003.5") if not $engvallex->is_valid_frame($frame);

      my $f_lemma = $engvallex->word_lemma($engvallex->frame_word($frame));
      $f_lemma =~ s/ /_/g;
      push @output, "valency003.1-$f_lemma:" . $node->{t_lemma} if lc($f_lemma) ne lc($node->{t_lemma});

      my @mem = map { s/\(.*?\)//g ; $_ }
        grep $_ && !m/EMPTY|---/
          ,split / /,$engvallex->serialize_frame($frame);
      my @mem_oblig = grep !m/^\?/,@mem;
	  my %ech_func;
	  for my $ech (GetEChildren($node))
	  {
		  if ($ech->{functor} eq "SM"){
		  	for my $smech (GetEChildren($ech)){
				$ech_func{$smech->{functor}}++;
			}
		  }
		  else{
			  $ech_func{$ech->{functor}}++;
		  }
	  }
      foreach my $mem (@mem_oblig)
	  {
		  # Pripad, kdy je slot zaplnen disjunkci funktoru. Napr. MANN|EXT
		  if ($mem =~ m/\|/)
		  {
			  my @disj = split /\|/, $mem;
			  my $found = 0;
			  foreach my $disj (@disj)
			  {
				  $found = 1 if defined $ech_func{$disj};
			  }
			  push @output, "valency003.2-$mem" if not $found;
		  }
		  else
		  {
		  	push @output, "valency003.2-$mem" if not defined $ech_func{$mem};
		  }
      }
      foreach my $ch (grep /^(?:ACT|PAT|ADDR|EFF|ORIG|[DC]PHR)$/, map {$_->{functor}} GetEChildren($node))
	  {
        push @output, "valency003.3-$ch"
          unless first {/\b$ch\b/} @mem;
      }
  
	  return @output;
}

sub valency004 {
  my ($tree, $engvallex, $prefix) = @_;
  my @output;
  my %lex;
  my $lex;

  my $node = $tree;
    while ($node) {
      if(! $node->{is_generated} and $lex = $node->attr('a/lex.rf')){
        $lex{$lex} = $node;
      }
    } continue {
      $node = $node->following;
    }

    $node = $tree;
	while ($node) {
	  next if grep { $_->{type} =~ m/^Engvallex/ } ListV($node->{annot_comment});
      my $error1; # test 2 will be skipped if error 1 is found
      if($node->{is_generated} and $lex = $node->attr('a/lex.rf')){
        unless ($node->{'val_frame.rf'} eq $lex{$lex}->{'val_frame.rf'}){
		  push @output, "valency004.1-warning $node->{id}" if ($node->{t_lemma} !~ m/^(?:be|have|can|must|may|might)$/);
          $error1 = 1;
        }
      }
      if ( ! $error1 and ! $node->{"val_frame.rf"}
           and $node->attr("a/lex.rf")
           and PML_T::GetANodeByID($node->attr("a/lex.rf"))->attr("m/tag") =~ m/^V/
           and $node->{functor} ne 'DPHR'
         ){
        if($node->{is_generated} and $lex = $node->attr('a/lex.rf')
           and ! $lex{$lex}->{'val_frame.rf'}){
          push @output, "valency004.2-warning $node->{id}"
        }
      }
    } continue {
      $node = $node->following;
    }
  return @output;
}

my @list_of_node_checks = (\&{'coord001'}, \&{'coord002'}, \&{'coord003'}, \&{'coord004'}, \&{'coord005'}, \&{'coord006'},
	        			   \&{'coref_gram001'},
						    \&{'bracket001'},
						   \&{'links002'}, \&{'links003'},
						   \&{'structure001'}, \&{'structure002'}, \&{'structure003'}, \&{'structure004'}, \&{'structure005'}, \&{'structure006'},
						   \&{'structure007'}, \&{'structure008'}, \&{'structure009'}, \&{'structure010'}, \&{'structure014'}, \&{'structure015'},
						   \&{'structure017'}, \&{'structure018'}, \&{'structure019'}, \&{'structure020'}, \&{'structure021'}, \&{'structure022'},
					       \&{'structure023'}, \&{'structure026'}, \&{'structure027'}, \&{'structure028'}, \&{'structure029'}, \&{'structure032'},
						   \&{'structure033'}, \&{'structure034'}, \&{'structure035'}, \&{'structure036'}, \&{'structure037'}, 
						   \&{'t_lemma001'},
						   \&{'valency001'},\&{'valency002'}, \&{'valency003'});
my @list_of_tree_checks = (\&{'structure030'}, \&{'structure031'}, \&{'valency004'});
my @list_of_file_checks = (\&{'links001'});

sub check_node
{
	my ($node, $engvallex, $prefix) = @_;
	$node = $this if not ref($node);
	my $OK = 1;

	if (not ref($engvallex))
	{
		my $file;
		($prefix,$file) = FindVallex();
		$engvallex = TrEd::EngValLex::NonGUI->new($file);
	}
	
	delete $node->{anot_error};

	for my $check (@list_of_node_checks)
	{
		my @response = &$check($node, $engvallex, $prefix);
		for my $response (@response)
		{
			next if grep { /^$response$/ } @{$node->{ignored_errors}};
			AddToListUniq($node, 'anot_error', $response);
			$OK = 0;
			#my $treeno = $node->{id};
			#$treeno =~ s/.*-s([0-9]+).*/\1/;
			my $treeno = TredMacro::ThisAddress($node);
			$treeno =~ s/.*#([0-9]+)\..*/\1/;
			TredMacro::GotoTree($treeno);
			$this = $node;
		}
	}

	return $OK;
}

sub check_tree
{
	my ($tree, $engvallex, $prefix) = @_;
	$tree = $root if not ref($tree);
	my $OK = 1;

	if (not ref($engvallex))
	{
		my $file;
		($prefix,$file) = FindVallex();
		$engvallex = TrEd::EngValLex::NonGUI->new($file);
	}
	
	my @nodes = reverse ($tree, $tree->descendants);
	for my $node (@nodes)
	{
		$OK = 0 if not check_node($node, $engvallex, $prefix);
	}

	for my $check (@list_of_tree_checks)
	{
		my @response = &$check($tree, $engvallex, $prefix);
		for my $response (@response)
		{
			my ($error, @nodes) = split(" ", $response);
			@nodes = map { PML::GetNodeByID($_) } @nodes;
			for my $node (@nodes)
			{
				next if grep { /^$error$/ } @{$node->{ignored_errors}};
				$OK = 0;
				AddToListUniq($node, 'anot_error', $error);
				#		my $treeno = $node->{id};
				# $treeno =~ s/.*-s([0-9]+).*/\1/;
				my $treeno = TredMacro::ThisAddress($node);
				$treeno =~ s/.*#([0-9]+)\..*/\1/;
				TredMacro::GotoTree($treeno);
				$this = $node;
			}
		}
	}

	return $OK;
}

sub check_file
{
	my $no_gui = shift;
	my @trees = reverse GetTrees();
	my $OK = 1;

	my ($prefix,$file) = FindVallex();
	my $engvallex = TrEd::EngValLex::NonGUI->new($file);

	for my $tree (@trees)
	{
		$OK = 0 if not check_tree($tree, $engvallex, $prefix);
	}

	for my $check (@list_of_file_checks)
	{
		#store the current position
		my ($cur_node, $cur_tree) = ($this, CurrentTreeNumber() + 1);

		#check the file - the position can be changed
		my @response = &$check($engvallex, $prefix, @trees);

		#back to stored position
		TredMacro::GotoTree($cur_tree);
		$this = $cur_node;

		for my $response (@response)
		{
			my ($error, @nodes) = split(" ", $response);
			@nodes = map { PML::GetNodeByID($_) } @nodes;
			for my $node (@nodes)
			{
				next if grep { /^$error$/ } @{$node->{ignored_errors}};
				$OK = 0;
				AddToListUniq($node, 'anot_error', $error);
				#my $treeno = $node->{id};
				#$treeno =~ s/.*-s([0-9]+).*/\1/;
				my $treeno = TredMacro::ThisAddress($node);
				$treeno =~ s/.*#([0-9]+)\..*/\1/;
				TredMacro::GotoTree($treeno);
				$this = $node;
			}
		}
	}

	if ($no_gui != 1)
	{
		if ($OK) { ToplevelFrame->messageBox(-message=>"No errors found",-type=>'ok') }
	}
}

sub ignore_current_error
{
	my @errors = ListV($this->{anot_error});
	return unless scalar(@errors);
	my $error = shift @errors;
	push @{$this->{ignored_errors}}, $error;
	$this->{anot_error} = List(@errors);
}

sub goto_next_error
{
	my $first_tree = CurrentTreeNumber();
	my $current_tree = $first_tree;
	my @trees = GetTrees();
	my $node = $this;
	my $found = 0;
	my $continue = 1;
	
	if (!$node->following)
	{
		if ( $current_tree < $#trees ) { $node = $trees[++$current_tree]; }
		else { $continue = 0; }
	}
	else { $node = $node->following; }

	while ( not $found and $continue )
	{
		$found = 1 if scalar(ListV($node->{anot_error}));

		if ($found)
		{
			TredMacro::GotoTree($current_tree + 1) if $current_tree != $first_tree;
			$this = $node;
			return;
		}
		else
		{
			if (!$node->following)
			{
				if ( $current_tree < $#trees ) { $node = $trees[++$current_tree]; }
				else { $continue = 0; }
			}
			else { $node = $node->following; }
		}
	}

	ToplevelFrame->messageBox(-message=>"No error in the rest of the file",-type=>'ok');
}

