# -*- cperl -*-

#ifndef PML_A_Anot
#define PML_A_Anot

#include <contrib/pml/PML_A.mak>

#encoding iso-8859-2

package PML_A_Anot;

BEGIN { import PML_A; }

=pod

=head1 PML_A_Anot

PML_T_Anot.mak - Miscellaneous macros for annotating the
analytical layer in the way of Prague Dependency Treebank (PDT)
2.0.

=over 4

=cut

#key-binding-adopt PML_A_Edit
#menu-binding-adopt PML_A_Edit

#include <contrib/support/unbind_edit.inc>

#unbind-key a
#unbind-key L
#unbind-key T
#unbind-key m
#unbind-key p
#unbind-key +
#unbind-key CTRL+A
#unbind-key CTRL+a
#unbind-key minus
#unbind-key KP_Subtract
#unbind-key CTRL+plus
#unbind-key CTRL+KP_Add

#ifdef anot_only
#unbind-key CTRL+R
#endif anot_only

#remove-menu Edit afun
#remove-menu Add This to a/aux.rf of Marked Node
#remove-menu Add This to a/lex.rf of Marked Node
#remove-menu Auto-assign afun to all nodes without afun in the tree
#remove-menu Auto-assign afun to the current node
#remove-menu Display tectogrammatical tree
#remove-menu Edit morphological lemma
#remove-menu Edit morphological tag
#remove-menu Remove This from a/*.rf of Marked Node

sub CreateStylesheets{
  SetStylesheetPatterns(<<'EOF','PML_A_Anot',1);
node:<? $${afun} eq "AuxS" ? '${id}' : '${m/form}' ?><?
  $${annot_comment} ?
    join('',' ',map{
      if(/M-l/){'#{#506aff}M'}elsif(/M-t/){'#{#506aff}t'}
      elsif(/Sent/){'#{brown}s'}elsif(/Word/){'#{brown}w'}
      elsif(/Typo/){'#{#506aff}!'}elsif(/Re/){'#{red}X'}
      elsif(/Other/){'#{red}?'}else{'#{red}!'}
    }map{$_->{type}} ListV($this->{annot_comment}))
  :'' ?>

style:<? my $node = $this;
         $node = (first {$_->{afun} !~ /^(?:Aux[CP])$/} $this->children) || $node if $${afun} =~ /^(?:Aux[CP])$/o;
         if($node->{afun} !~ /^(?:Aux[GYX])$/o
            and $this->parent
            and $this->parent->{afun} =~ /^(?:Coord|Apos)$/
            and ! $node->{is_member}){
  '#{Line-fill:green}'}else{''}
?>

node:<?
(($${afun} eq q{}) or
($${afun}=~/^Apos|^Coord/ and PML_A::ExpandCoord($this)==0))
 ? '#{customerror}!!'
 : '#{customafun}'
?>${afun}<?
  if ($${is_member}) {
    my $p=$this->parent;
    $p=$p->parent while $p and $p->{afun}=~/^Aux[CP]$/;
    ($p and $p->{afun}=~/^(Ap)os|(Co)ord/ ? "_#{customcoappa}\${is_member=$1$2}" : "_#{customerror}\${is_member=ERR}")
  } else { "" }
?><? $${is_parenthesis_root} ? '#{customcoappa}_${is_parenthesis_root=Pa}' : '' ?>
hint:<? join "\n",map{'! '.$_->{type}.':'.$_->{text}}ListV($this->{annot_comment})?>
EOF
}#CreateStylesheets

sub get_value_line_hook {
  my @sent;
  my $node = $root;
  while ($node = $node->following) {
    push @sent,$node;
  }
  my @out;
  foreach $node (sort { $a->{ord} <=> $b->{ord} } @sent){
    my $token = join(" ",map { $_->{token} } ListV($node->attr('m/w')));
    my $form = $node->attr('m/form');
    if ($form ne $token){
      push @out,(['['.$token.']',$node,'-over=>1','-foreground=>'.CustomColor('spell')]);
    }
    push@out,([$form,$node]);
    unless($node->attr('m/w/no_space_after')){
      push@out,([" ","space"]);
    }
  }
  if($root->{eng_sentence}){
    push@out,(["\n".$root->{eng_sentence},'-foreground=>darkgreen']);
  }
  return \@out;
} # get_value_line_hook

sub switch_context_hook {
  CreateStylesheets();
  SetCurrentStylesheet('PML_A_Anot');
  Redraw() if GUI();
  undef$PML::arf;
}

#bind OpenValFrameList CTRL+Return menu Show valency lexicon entry for the current word
sub OpenValFrameList {
  shift unless @_ and ref($_[0]);
  my $node = shift || $this;
  my %opts = @_;

  my $vallex_file = 'vallex2.xml';
  $vallex_file = ResolvePath(FileName(),$vallex_file,1);

  local $ValLex::GUI::frameid_attr=undef;
  local $ValLex::GUI::lemma_attr=undef;
  local $ValLex::GUI::framere_attr=undef;
  local $ValLex::GUI::sempos_attr=undef;
  warn "vallex2.xml should be opened!";
  my $pos = substr($node->attr('m/tag'),0,1);
  return unless ($pos=~/[NVAD]/);

  my $lemma = $node->attr('m/lemma');

  # this is specific for czech
  $lemma =~ s/[_\`&].*$//;
  my $se = first { $_->{afun} eq 'AuxT' } GetEChildren($node,\&DiveAuxCP);
  if ($se) {
    $lemma .= $se->attr('m/tag')=~/^P7-X3/ ? '_si' : '_se';
  }

  ValLex::GUI::ChooseFrame({
    -vallex_file => $vallex_file,
    -lemma => $lemma,
    -pos => $pos,
    -assignfunc => sub{},
    %opts
   });
  ChangingFile(0);
}

#bind AddComment ! menu Add Comment
sub AddComment {
  my @list = $this->type('annot_comment/type/')->get_values();
  my $dialog = [$list[0]];
  ListQuery('Comment type',
            'browse',
            \@list,
            $dialog) or return;

  # For Re-generate, prepare the old sentence to be edited.
  my $default_text;
  if ($dialog->[0] eq 'Re-generate' and $this == $root){
    $default_text = GetSentenceString($this);
    $default_text =~ s/ *$//;
  }
  my $text = QueryString('Comment text','Text:',$default_text);
  return unless defined $text;
  my%comment=(type=>$dialog->[0],
              text=>$text);
  AddToList($this,'annot_comment',\%comment);
}#AddComment

#bind EditComment ? menu Edit Comment
sub EditComment{
  ChangingFile(EditAttribute($this,'annot_comment'));
}#EditComment



sub enable_edit_node_hook { 'stop' }

sub enable_attr_hook {
  if ($_[0] =~ m=^(?:a(?:fun|nnot_comment(?:|/t(?:ext|ype)))|is_(?:member|parenthesis_root))$=){
    1 } else {'stop'}
} # enable_attr_hook

sub AfunAssign{
  $this->{afun}=$_[0] unless $this==$root;
  $this = $this->following;
}

#bind AssignPred to q menu AssignPred
sub AssignPred {
  AfunAssign('Pred');
}
#bind AssignPnom to n menu AssignPnom
sub AssignPnom {
  AfunAssign('Pnom');
}
#bind AssignAuxV to v menu AssignAuxV
sub AssignAuxV {
  AfunAssign('AuxV');
}
#bind AssignSb to s menu AssignSb
sub AssignSb {
  AfunAssign('Sb');
}
#bind AssignObj to b menu AssignObj
sub AssignObj {
  AfunAssign('Obj');
}
#bind AssignAtr to a menu AssignAtr
sub AssignAtr {
  AfunAssign('Atr')
}
#bind AssignAdv to d menu AssignAdv
sub AssignAdv {
  AfunAssign('Adv');
}
#bind AssignCoord to i menu AssignCoord
sub AssignCoord {
  AfunAssign('Coord');
}
#bind AssignAuxT to t menu AssignAuxT
sub AssignAuxT {
  AfunAssign('AuxT');
}
#bind AssignAuxR to r menu AssignAuxR
sub AssignAuxR {
  AfunAssign('AuxR');
}
#bind AssignAuxP to p menu AssignAuxP
sub AssignAuxP {
  AfunAssign('AuxP');
}
#bind AssignApos to u menu AssignApos
sub AssignApos {
  AfunAssign('Apos');
}
#bind AssignAuxC to c menu AssignAuxC
sub AssignAuxC {
  AfunAssign('AuxC');
}
#bind AssignAuxO to o menu AssignAuxO
sub AssignAuxO {
  AfunAssign('AuxO');
}
#bind AssignAtv to h menu AssignAtv
sub AssignAtv {
  AfunAssign('Atv');
}
#bind AssignAtvV to j menu AssignAtvV
sub AssignAtvV {
  AfunAssign('AtvV');
}
#bind AssignAuxZ to z menu AssignAuxZ
sub AssignAuxZ {
  AfunAssign('AuxZ');
}
#bind AssignAuxY to y menu AssignAuxY
sub AssignAuxY {
  AfunAssign('AuxY');
}
#bind AssignAuxG to g menu AssignAuxG
sub AssignAuxG {
  AfunAssign('AuxG');
}
#bind AssignAuxK to k menu AssignAuxK
sub AssignAuxK {
  AfunAssign('AuxK');
}
#bind AssignAuxX to x menu AssignAuxX
sub AssignAuxX {
  AfunAssign('AuxX');
}
#bind AssignExD to e menu AssignExD
sub AssignExD {
  AfunAssign('ExD');
}

#bind RotateMember to 1 menu Change is_member
sub RotateMember{
  $this->{is_member} =! $this->{is_member};
}

#bind RotateParenthesisRoot to 2 menu Change is_parenthesis_root
sub RotateParenthesisRoot{
  $this->{is_parenthesis_root} =! $this->{is_parenthesis_root};
}

1;

=back

=cut

#endif
